/*
    DPS - Dynamic Parallel Schedules
    Copyright (C) 2000-2003 Sebastian Gerlach

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
	DPS Layer 1 - Base library

	This file is part of the base namespace, containing all the basic 
	functionality used by DPS. It provides support for reflection, smart
	pointers, serialization, and containers. It also contains cross-platform
	abstractions for several useful system services, such as threads, sockets,
	and synchronization primitives.
	The classes in this namespace have no dependencies on other DPS classes.
*/

/*!	\file base.h
	\brief Base classes

	Contains base classes for reflection and garbage collection, and associated
	macros. Includes all other base-related header files.
*/

#ifndef DPS_BASE_H_INCLUDED
#define DPS_BASE_H_INCLUDED

//
// Version number
//
#define DPS_VERSION_MAJOR	0
#define DPS_VERSION_MINOR	4
#define DPS_VERSION_BUILD	2


//
// Simple compiler and platform detection
//
#if defined(_MSC_VER)
// This is the microsoft C compiler, we are thus compiling for Win32
#define DPS_PLATFORM_WIN32
#if _MSC_VER >= 1310
#define DPS_COMPILER_GCC
#else
#define DPS_COMPILER_MS
#endif

#elif defined(WIN32)
// This is mingw32 
#define DPS_PLATFORM_WIN32
#define DPS_COMPILER_GCC

#else
// This is plain UNIX
#define DPS_PLATFORM_POSIX
#define DPS_COMPILER_GCC
#endif

#ifdef _MSC_VER // Cannot use COMPILER_MS here, since 13.10 has COMPILER_GCC defined
// Some of these must be defined in advance of including the standard
// header files in order to avoid those pesky warnings about excessively
// long identifiers (which this code generates in large quantities).
#pragma warning(disable:4786)	// identifier name was truncated to 'x' characters
#pragma warning(disable:4503)	// decorated name length exceeded, name was truncated
#pragma warning(disable:4355)	// 'this' used in base member initializer list
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
//#include <wchar.h>
#include <stdarg.h>

#include <vector>
#include <string>
#include <map>
#include <algorithm>

#ifdef DPS_PLATFORM_WIN32
// Windows specific include files
#ifndef _WIN32_WINNT
// If there is no previous WINNT version defined, define at least NT 4.0
#define _WIN32_WINNT 0x0400
#endif
#include <windows.h>
#include <io.h>
#include <direct.h>
#include <malloc.h>
// Undefine min and max (these should not be macros anyway)
#undef min
#undef max

#if _MSC_VER<1300
// For older MS compilers, set min and max to _MIN and _MAX
#define min _MIN
#define max _MAX
#endif
#else
#define stricmp strcasecmp
#endif

//
// Check for debug version
//
#ifdef _DEBUG
#define DPS_DEBUG
#endif

#if defined(DPS_DEBUG)
#ifdef DPS_COMPILER_MS
//! Call to debugger
static inline void callInt3() 
{ 
	__asm int 3;
}
#else
//! Call to debugger
static inline void callInt3()
{
}
#endif
#define debugBreak callInt3()
#define check(x,y) { if(!(x)) { printf y; debugBreak; } }
#define DBGDUMP(x) dump x
#else
#define check(x,y)
#define DBGDUMP(x)
#endif

#if defined(__GLIBCPP__)
#include <bits/atomicity.h>
#endif

//! 8 bit signed integer
typedef char Int8;
//! 16 bit signed integer
typedef short Int16;
//! 32 bit signed integer
typedef int Int32;
//! 8 bit unsigned integer
typedef unsigned char UInt8;
//! 16 bit unsigned integer
typedef unsigned short UInt16;
//! 32 bit unsigned integer
typedef unsigned int UInt32;
//! 32 bit floating point number
typedef float Float;
//! 64 bit floating point number
typedef double Double;
//! Integer for sizes
typedef size_t Size;
//! Integer for cross-platform sizes 
/*! Since size_t might be 32 or 64 bits (or maybe even something completely 
	different), we have to decide what to use when serializing data. For 
	performance reasons, we use 32 bits here.
*/
typedef UInt32 XPSize;

//! Single-byte character type
typedef char AnsiChar;
//! Unicode (two-byte) character type
typedef wchar_t UnicodeChar;

typedef bool Bool;

#ifdef _MSC_VER
//! 64 bit signed integer
typedef __int64 Int64;
//! 64 bit unsigned integer
typedef unsigned __int64 UInt64;
//! 64 bit signed integer format flag for printf
#define I64F "%I64d"
//! 64 bit unsigned integer format flag for printf
#define U64F "%I64u"
#else
//! 64 bit signed integer
typedef long long Int64;
//! 64 bit unsigned integer
typedef unsigned long long UInt64;
//! 64 bit signed integer format flag for printf
#define I64F "%lld"
//! 64 bit unsigned integer format flag for printf
#define U64F "%llu"
#endif

typedef UInt32 Result;

namespace base
{
	void fatalError(const char *fmt, ...);
};


// Include synchronization classes
#include "basesync.h"

// Include type classes
#include "basetypes.h"

//! Contains utility classes
/*! This namespace contains the functions for the smart pointers, the serializer, the
	custom containers, and some other stuff.
*/
namespace base
{
	extern std::string executableName;

	class Object;

	class BinarySerializationTarget;
	class BinarySerializationSource;
	class InputStream;
	class OutputStream;

	/*! \brief Smart pointer class
		\ingroup user
		This is a simple smart pointer class used in the DPS smart pointer model. It uses
		in-place reference counting (e.g. the reference count is stored in the pointed
		objects.
	*/
	template <typename T> class Ptr 
	{
	private:
		//! The pointer
		T* p;
		
	public:
		//! Constructor
		Ptr() { p=NULL; }
		//! Constructor from pointer
		Ptr(T* np) { p=np; if(p) p->upcount(); }
		//! Constructor from Ptr
		Ptr(const Ptr<T>& np) { p=np; if(p) p->upcount(); }
		//! Destructor
		~Ptr(void) { if(p) p->downcount(); }
		
		//! Cast to contained type
		operator T*(void) const { return p; }
		//! Cast to reference of contained type
		T& operator*(void) { return *p; }
		//! Member operator
		T* operator->(void) const { return p; }
		//! Assignment from Ptr
		Ptr& operator=(const Ptr<T> &np) { T* op=p; p = np.p; if(p) p->upcount(); if(op) op->downcount(); return *this; }
		//! Assignment from pointer
		Ptr& operator=(T * const np) { T* op=p; p=np; if(p) p->upcount(); if(op) op->downcount(); return *this; }
		//! Equality to pointer
		bool operator==(const T* op) const { return p==op; }
		//! Equality to Ptr
		bool operator==(const Ptr<T>& op) const { return p==op.p; }

		//! Serialize pointed object
		void write(OutputStream *s) const
		{
			s->writeRef(p);
		}
		//! Deserialize pointed object
		void read(InputStream *s)
		{
			p=(T*)s->readRef();
			if(p)
				p->upcount();
		}
	};

	/*! \brief Const smart pointer class
		\ingroup user
		This is a simple smart pointer class used in the DPS smart pointer model. It uses
		in-place reference counting (e.g. the reference count is stored in the pointed
		objects.
	*/
	template <typename T> class ConstPtr 
	{
	private:
		//! The pointer
		const T* p;
		
	public:
		//! Constructor
		ConstPtr() { p=NULL; }
		//! Constructor from pointer
		ConstPtr(T* np) { p=np; if(p) p->upcount(); }
		//! Constructor from Ptr
		ConstPtr(const ConstPtr<T>& np) { p=np; if(p) p->upcount(); }
		//! Destructor
		~ConstPtr(void) { if(p) p->downcount(); }
		
		//! Cast to contained type
		operator T*(void) const { return p; }
		//! Cast to reference of contained type
		const T& operator*(void) const { return *p; }
		//! Member operator
		const T* operator->(void) const { return p; }
		//! Assignment from Ptr
		ConstPtr& operator=(const ConstPtr<T> &np) { T* op=p; p = np.p; if(p) p->upcount(); if(op) op->downcount(); return *this; }
		//! Assignment from pointer
		ConstPtr& operator=(const T * const np) { const T* op=p; p=np; if(p) p->upcount(); if(op) op->downcount(); return *this; }
		//! Equality to pointer
		bool operator==(const T* op) const { return p==op; }
		//! Equality to Ptr
		bool operator==(const Ptr<T>& op) const { return p==op.p; }

		//! Serialize pointed object
		void write(OutputStream *s) const
		{
			s->writeRef(p);
		}
		//! Deserialize pointed object
		void read(InputStream *s)
		{
			p=(T*)s->readRef();
			if(p)
				p->upcount();
		}
	};

	/*! \brief Smart pointer class without refcount
		\ingroup user
		This is a simple smart pointer class used in the DPS smart pointer model. It does 
		not use reference counting, and should be used when a reverse reference is
		required (e.g. the parent-to-child pointer is a Ptr, the child-to-parent point is
		a PtrNoRef. In this case the structure can still be collected since it is not
		circular, and the reference to the parent is still serialized).
	*/
	template <typename T> class PtrNoRef
	{
	private:
		//! The pointer
		T* p;
		
	public:
		//! Constructor
		PtrNoRef() { p=NULL; }
		//! Constructor from pointer
		PtrNoRef(T* np) { p=np;  }
		
		//! Cast to contained type
		operator T*(void) const { return p; }
		//! Cast to reference of contained type
		T& operator*(void) const { return *p; }
		//! Member operator
		T* operator->(void) const { return p; }
		//! Assignment from pointer
		PtrNoRef& operator=(T * const np) { p=np; return *this; }
		//! Equality to pointer
		bool operator==(const T* op) const { return p==op; }

		//! Serialize pointed object
		void write(OutputStream *s) const { s->writeRef(p); }
		//! Deserialize pointed object
		void read(InputStream *s) { p=(T*)s->readRef(); }
	};

	/*! \brief Constant smart pointer class without refcount
		\ingroup user
		This is a simple smart pointer class used in the DPS smart pointer model. It does 
		not use reference counting, and should be used when a reverse reference is
		required (e.g. the parent-to-child pointer is a Ptr, the child-to-parent point is
		a PtrNoRef. In this case the structure can still be collected since it is not
		circular, and the reference to the parent is still serialized).
	*/
	template <typename T> class ConstPtrNoRef
	{
	private:
		//! The pointer
		const T* p;
		
	public:
		//! Constructor
		ConstPtrNoRef() { p=NULL; }
		//! Constructor from pointer
		ConstPtrNoRef(const T* np) { p=np;  }
		
		//! Cast to contained type
		operator const T*(void) const { return p; }
		//! Cast to reference of contained type
		const T& operator*(void) const { return *p; }
		//! Member operator
		const T* operator->(void) const { return p; }
		//! Assignment from pointer
		ConstPtrNoRef& operator=(T * const np) { p=np; return *this; }
		//! Equality to pointer
		bool operator==(const T* op) const { return p==op; }

		//! Serialize pointed object
		void write(OutputStream *s) const { s->writeRef(p); }
		//! Deserialize pointed object
		void read(InputStream *s) { p=(const T*)s->readRef(); }
	};

	/*! \brief Derived class from std::vector to add serialization features
		\ingroup internal
		This class is identical to std::vector, except for the added serialization
		methods. It should not exist, but is required to compensate for the lack
		of partial template specialization in some compilers.
	*/
	template <typename T> class vector : public std::vector<T>
	{
	public:
		vector() : std::vector<T>() { }
		vector(const vector& org) : std::vector<T>(org) { }
		vector& operator=(const vector& o) { (*(std::vector<T>*)this)=o; return *this;}
		void write(OutputStream *s) const
		{ 
			XPSize ws=(XPSize)size();
			typehandler::TypeHandler<XPSize>::write(s,ws);
			for(typename std::vector<T>::const_iterator i=begin();i!=end();i++)
				typehandler::TypeHandler<T>::write(s,*i);
		}
		void read(InputStream *s) 
		{ 
			XPSize ws;
			typehandler::TypeHandler<XPSize>::read(s,ws);
			for(Size i=0;i<ws;i++)
			{
				T t;
				typehandler::TypeHandler<T>::read(s,t);
				push_back(t);
			}
		} 
	};

	//! Contains internal templated functions for byte swapping
	/*!	This namespace contains multiple specialized function implementations
		for byte swapping. This is used by the serializer when mixed-endian
		machines are communicating together.
	*/
	namespace byteswapper
	{
		/*! \brief Generic byte swap (internal)
			\ingroup internal
			\param data Pointer to data to byteswap
			Swaps the first size bytes of data.
		*/
		template<int size> static inline void swapp(void *data)
		{
			UInt8 dc[size];
			UInt32 p;
			for(p=0;p<size;p++)
				dc[p]=((UInt8*)data)[p];
			for(p=0;p<size;p++)
				((UInt8*)data)[p]=dc[size-1-p];
		}

		//! Swap bytes for 8-bit word (internal)
		/*! \ingroup internal
			\param data Pointer to data to byteswap
			Do nothing, since there is only one byte...
		*/
		template<> static inline void swapp<1>(void *data)
		{
		}
		//! Swap bytes for 16-bit word (internal)
		/*! \ingroup internal
			\param data Pointer to data to byteswap
			Swaps the first 2 bytes of data.
		*/
		template<> static inline void swapp<2>(void *data)
		{
			UInt16 a=*(UInt16*)data;
			*(UInt16*)data=
				((a&0x00ff)<< 8)|
				((a>> 8)&0x00ff);
		}
		//! Swap bytes for 32-bit word (internal)
		/*! \ingroup internal
			\param data Pointer to data to byteswap
			Swaps the first 4 bytes of data.
		*/
		template<> static inline void swapp<4>(void *data)
		{
			UInt32 a=*(UInt32*)data;
			*(UInt32*)data=
				((a&0x000000ff)<<24)|
				((a&0x0000ff00)<< 8)|
				((a>> 8)&0x0000ff00)|
				((a>>24)&0x000000ff);
		}
		//! Swap bytes for 64-bit word (internal)
		/*! \ingroup internal
			\param data Pointer to data to byteswap
			Swaps the first 8 bytes of data.
		*/
		template<> static inline void swapp<8>(void *data)
		{
			UInt64 a=*(UInt64*)data;
#ifndef DPS_COMPILER_GCC
			*(UInt64*)data=
				((a&0x00000000000000ffi64)<<56)|
				((a&0x000000000000ff00i64)<<40)|
				((a&0x0000000000ff0000i64)<<24)|
				((a&0x00000000ff000000i64)<< 8)|
				((a>> 8)&0x00000000ff000000i64)|
				((a>>24)&0x0000000000ff0000i64)|
				((a>>40)&0x000000000000ff00i64)|
				((a>>56)&0x00000000000000ffi64);
#else
			*(UInt64*)data=
				((a&0x00000000000000ffLL)<<56)|
				((a&0x000000000000ff00LL)<<40)|
				((a&0x0000000000ff0000LL)<<24)|
				((a&0x00000000ff000000LL)<< 8)|
				((a>> 8)&0x00000000ff000000LL)|
				((a>>24)&0x0000000000ff0000LL)|
				((a>>40)&0x000000000000ff00LL)|
				((a>>56)&0x00000000000000ffLL);
#endif
		}

		//! Generic byte swap function
		/*! \ingroup internal
			This function should be called whenever bytes have to be swapped.
			It will automagically forward the call to the appropriate internal
			swapping function according to sizeof(item).
			\param item Item to byte-swap.
		*/
		template<typename t> static void swap(t& item)
		{
			byteswapper::swapp<sizeof(t)>(&item);
		}
	}

//! Macro for transforming a string into a class
/*! \ingroup internal
	This silly macro is required since it is not acceptable to pass a constant string
	pointer as argument to a template. So we just wrap the string pointer in a class.
*/
#define STRING(x) struct __string_##x { static inline const char* name() { return #x; } }

//! Macro for transforming a string into a class
/*! \ingroup internal
	This silly macro is required since it is not acceptable to pass a constant string
	pointer as argument to a template. So we just wrap the string pointer in a class.
	Identical to STRING, but the created class has a different name.
*/
#define STRINGB(x) struct __stringb_##x { static inline const char* name() { return #x; } }

	//! Class for holding items in a linked list of types
	/*! \ingroup internal
		Template parameters:
		- tc: Owner class
		- cName: Member name
		- serialName: Member serialization name
		- getter: Member serialization helper class
		- next: next member in stack
	*/
	template<typename tc, typename cName, typename serialName, typename getter, typename next> struct MemberStack
	{
		//! Next item
		typedef next nextItem;
		//! Current item type
		typedef typename getter::type type;

		//! Perform type comparison
		static inline bool isA(const AnsiChar *tn)
		{
			return next::isA(tn) || getter::isA(tn);
		}
		//! Write item to output stream
		static inline void write(OutputStream *s, const tc& base) 
		{
			next::write(s,base);
			ENTERITEM(serialName::name());
			typehandler::write(s,getter::getConst(base));
			LEAVEITEM(serialName::name());
		}
		//! Read item from input stream
		static inline void read(InputStream *s, tc& base) 
		{ 
			next::read(s,base);
			ENTERITEM(serialName::name());
			typehandler::read(s,getter::get(base));
			LEAVEITEM(serialName::name());
		}
	};
	
	//! Dummy class for last item in list
	/*! \ingroup internal
		Template parameters:
		- tc: Owner class
	*/
	template<typename tc> struct MemberStackTop
	{
		static inline bool isA(const AnsiChar * ) { return false; }
		static inline void write(OutputStream *, const tc& ) { }
		static inline void read(InputStream *, tc& ) { }
	};

#define MEMBERGETTER(t,n) struct getter##n { typedef t type; static inline t& get(containerType& item) { return item.n; } static inline const t& getConst(const containerType& item) { return item.n; } static inline bool isA(const AnsiChar *tn) { return false; } };
#define BASEGETTER(t,n) struct getter##n { typedef t type; static inline t& get(containerType& item) { return item; } static inline const t& getConst(const containerType& item) { return item; } static inline bool isA(const AnsiChar *tn) { return type::isA(tn,false); } };
//#define SERIAL(t,n) struct __serialize##n { static inline void iwrite(base::BinarySerializationTarget *trg, const containerType& a) { trg->write(a.n); } static inline void iread(base::BinarySerializationSource *src, containerType& a) { src->read(a.n); } };
#define COMMON(t,n,ren) STRING(n); STRINGB(ren); MEMBERGETTER(t,n); typedef base::MemberStack<containerType, __string_##n, __stringb_##ren, getter##n, __curt##n >

#define TYPENAMESTD(tc) public: static const AnsiChar *getTypeName() { return #tc; } 
#define TYPENAMETEMPLATE1(tc,p1) public: static const char *getTypeName() { RETSTRCAT3(#tc"<",base::typehandler::TypeHandler<p1>::name(),">"); }
#define TYPENAMETEMPLATE2(tc,p1,p2) public: static const char *getTypeName() { RETSTRCAT5(#tc"<",base::typehandler::TypeHandler<p1>::name(),",",base::typehandler::TypeHandler<p2>::name(),">"); }

#define IDENTIFYSTATIC(a) \
	typedef a containerType; \
	static Size getSize() { return sizeof(a); } \
	static bool isA(const AnsiChar *tn, bool exact) { if(getTypeName()==tn) return true; if(!exact) return __baseClasses::isA(tn); return false; }

#define IDENTIFYVIRTUAL(tc) \
	virtual const AnsiChar *getTypeNameV() const { return getTypeName(); } \
	virtual Size getSizeV() const { return getSize(); } \
	virtual bool isAV(const AnsiChar *tn, bool exact) const { return isA(tn,exact); } 

#define SERIALSTATIC(tc) IDENTIFYSTATIC(tc) \
	void write(base::OutputStream* s) const { __baseClasses::write(s,*this); __members::write(s,*this); } \
	void read(base::InputStream* s) { __baseClasses::read(s,*this); __members::read(s,*this); } 

#define SERIALVIRTUAL(tc) IDENTIFYVIRTUAL(tc) \
	virtual void writeV(base::OutputStream* s) const { this->write(s); } \
	virtual void readV(base::InputStream* s) { this->read(s); } 

#define REGISTERCOMMON(x) virtual Size getTypeIndexV() const { static base::RegistrarItemHolder<x > registerItem; return registerItem.getIndex(); }
#define REGISTERCLASS(x) REGISTERCOMMON(x)
#define REGISTERTEMPLATECLASS(x) REGISTERCOMMON(x)
/*
#ifdef _MSC_VER
// This is the neat solution, just specify the explicit template instantiation.
#define REGISTERCLASS(x) template class base::RegistrarItemHolder<x >;
#define REGISTERTEMPLATECLASS(x) template class base::RegistrarItemHolder<x >;
#else
// Not quite as neat. First part works for simple types, second for templated
#define REGISTERCLASS(x) static void yuck() { base::RegistrarItemHolder<x > a; } 
#define REGISTERTEMPLATECLASS(x) base::RegistrarItemHolder<x> __reg;
#endif
*/
#define SIMPLECLASSDEF(tc) TYPENAMESTD(tc) SERIALSTATIC(tc) \
	typedef base::MemberStackTop<containerType> __baseClasses; public: typedef __baseClasses baseClassAccessPoint; typedef base::MemberStackTop<containerType> 

#define CUSTOMSERIAL(a) TYPENAMESTD(a) SERIALVIRTUAL(a) IDENTIFYSTATIC(a) REGISTERCLASS(a)
#define IDENTIFY(a) TYPENAMESTD(a) IDENTIFYSTATIC(a) IDENTIFYVIRTUAL(a) REGISTERCLASS(a)
#define IDENTIFYNOREG(a) TYPENAMESTD(a) IDENTIFYSTATIC(a) IDENTIFYVIRTUAL(a)

#define CLASSDEFNOREG(tc) TYPENAMESTD(tc) SERIALSTATIC(tc) SERIALVIRTUAL(tc) \
	private: typedef base::MemberStackTop<containerType> 

#define CLASSDEF(tc) TYPENAMESTD(tc) SERIALSTATIC(tc) SERIALVIRTUAL(tc) REGISTERCLASS(tc) \
	private: typedef base::MemberStackTop<containerType> 

#define TEMPLATEDEF1(tc,t1) TYPENAMETEMPLATE1(tc,t1) SERIALSTATIC(tc) SERIALVIRTUAL(tc) REGISTERTEMPLATECLASS(tc) \
	private: typedef base::MemberStackTop<containerType> 
#define TEMPLATEDEF2(tc,t1,t2) TYPENAMETEMPLATE2(tc,t1,t2) SERIALSTATIC(tc) SERIALVIRTUAL(tc) REGISTERTEMPLATECLASS(tc) \
	private: typedef base::MemberStackTop<containerType> 

#define TEMPLATEDEFNOREG1(tc,t1) TYPENAMETEMPLATE1(tc,t1) SERIALSTATIC(tc) SERIALVIRTUAL(tc) \
	private: typedef base::MemberStackTop<containerType> 
#define TEMPLATEDEFNOREG2(tc,t1,t2) TYPENAMETEMPLATE2(tc,t1,t2) SERIALSTATIC(tc) SERIALVIRTUAL(tc) \
	private: typedef base::MemberStackTop<containerType> 

#define SINGLEBASECLASS(cn) BASEGETTER(cn,id); private: typedef base::MemberStack<containerType,base::typehandler::TypeHandler<cn >, base::typehandler::TypeHandler<cn >, getter##id, base::MemberStackTop<containerType> > __baseClasses; public: typedef __baseClasses baseClassAccessPoint; 

	// HACK: BASECLASS uses the __LINE__ macro. So you should never put two BASECLASS
	// macros on the same line... 
#define BASECLASS2(cn,id) __curb##id; BASEGETTER(cn,id); \
	private: typedef base::MemberStack<containerType,base::typehandler::TypeHandler<cn >, base::typehandler::TypeHandler<cn >, getter##id, __curb##id >
#define BASECLASS1(cn,id) BASECLASS2(cn,id)
#define BASECLASS(cn) BASECLASS1(cn,__LINE__)

#define MEMBERS  __baseClasses; \
	public: typedef __baseClasses baseClassAccessPoint; \
	private: typedef base::MemberStackTop<containerType> 
#define ITEM(t,n) __curt##n; public: t n; private: COMMON(t,n,n)
#define ITEMR(t,n,ren) __curt##n; public: t n; private: COMMON(t,n,ren)
#define ITEMD(t,n) __curt##n; private: COMMON(t,n,n)

#ifndef DPS_COMPILER_GCC
#define ARRAY(t,n,s) __curt##n; typedef base::Array<t,s> __arr##n; public: __arr##n n; COMMON(__arr##n,n,n)
#else
#define ARRAY(t,n,s) __curt##n; typedef t __arr##n[s]; public: __arr##n n; COMMON(__arr##n,n,n)
#endif

#define CLASSEND __members; public: enum { checkForBaseClasses = sizeof(__baseClasses) } 

#define isType(x) isAV(x::getTypeName(),true)
#define isDerivedType(x) isAV(x::getTypeName(),false)



	extern CriticalSection refLock;

	//! Interlocked increment
	/*! \ingroup internal
	*/
	inline Int32 incrementRefCount(Int32* rc)
	{
#if defined(THREAD_UNSAFE)
		return ++(*rc);
#elif defined(DPS_PLATFORM_WIN32)
		return InterlockedIncrement((LPLONG)rc);
#elif defined(__GLIBCPP__)
		return __exchange_and_add((_Atomic_word*)rc, 1)+1;
#else
		ScopedLock lock(refLock);
		return ++(*rc);
#endif
	}
	
	//! Interlocked decrement
	/*! \ingroup internal
	*/
	inline Int32 decrementRefCount(Int32* rc)
	{
#if defined(THREAD_UNSAFE)
		return --(*rc);
#elif defined(DPS_PLATFORM_WIN32)
		return InterlockedDecrement((LPLONG)rc);
#elif defined(__GLIBCPP__)
        return __exchange_and_add((_Atomic_word*)rc, -1)-1;
#else
		ScopedLock lock(refLock);
		return --(*rc);
#endif
	}

	class Subsystem;

	extern Subsystem *subsystemHead;

	//! Descriptor class for a DPS subsystem
	/*! \ingroup user
		This class describes a subsystem for the DPS debug dump system.
	*/
	class Subsystem
	{
	private:
		//! Name of subsystem
		std::string name;
		//! Current dump level for this subsystem
		Int32 dumpLevel;
	public:
		//! Next subsystem in list
		Subsystem *next;
	public:
		//! Constructor
		Subsystem(const char *n) { name=n; dumpLevel=1; next=subsystemHead; subsystemHead=this; }
		//! Get name of subsystem
		const char *getName() const { return name.c_str(); }
		//! Get dump level of subsystem
		Int32 getDumpLevel() const { return dumpLevel; }
		//! Set dump level
		void setDumpLevel(Int32 dl) { dumpLevel=dl; }
	};

	//! Define subsystem to which a class belongs
	/*! \ingroup user
	*/
#define DPS_SUBSYSTEM(x) virtual const base::Subsystem& dpsGetSubsystem() { return x; }

	//! Dump orderly string to standard output
	/*! \ingroup internal
	*/
	void baseDump(const Subsystem& subsystem, Int32 level, const AnsiChar *owner, const AnsiChar *format, ...);

#ifdef DPS_PLATFORM_WIN32
	//! Dump error message
	void dumpError(int err);
#endif

	//! Default subsystem
	extern Subsystem baseSubsystem;

	//! Items stored in registrar
	/*! \ingroup internal
	*/
	class RegistrarItem
	{
	public:
		//! Next item in list
		RegistrarItem *next;
		//! Item name
		const AnsiChar *name;
		//! Item index
		Size index;

	public:
		//! Constructor
		RegistrarItem(const AnsiChar *name);

		//! Create an object of this registered type
		virtual Ptr<Object> create() = 0;
	};

	//! Specialized item stored in registrar
	/*! \ingroup internal
	*/
	template<typename regc> class RegistrarItemSpec : public RegistrarItem
	{
	public:
		//! Constructor
		RegistrarItemSpec() : RegistrarItem(regc::getTypeName()) { } 

		//! Create an object of this registered type
		virtual Ptr<Object> create() { return new regc; }

		//! Get index in Registrar
		Size getIndex() { return index; }
	};

	//! Holder for registrar item singleton
	/*! \ingroup internal
	*/
	template<typename regc> class RegistrarItemHolder
	{
		//! Held registrar item
		static RegistrarItemSpec<regc> item;
	public:
		//! Constructor
		RegistrarItemHolder() { /*item;*/ }

		//! Get index in Registrar
		static Size getIndex() { return item.getIndex(); }
	};
	//! Registrar singletons
	/*! \ingroup internal
	*/
	template<typename regc> RegistrarItemSpec<regc> RegistrarItemHolder<regc>::item;

	
	//! Base class for all objects that can be collected and identified
	/*! \ingroup user
		Deriving from this class is not a requirement for serialization
		of inline elements, but is required for pointed elements (since
		identification ability is required for class factory functionality).
		Any item that derives from Object is suitable for use with the
		Ptr smart pointer template.
	*/
	class Object
	{
		CLASSDEF(Object)
		MEMBERS
		CLASSEND;

	private:
		//! Reference count
		mutable Int32 crefs;
		
	public:
		//! Constructor
		Object() { crefs = 0; }
		//! Empty virtual destructor
		virtual ~Object() { }
		//! Increase reference count
		void upcount(void) const  
		{
			incrementRefCount(&crefs);
		}
		//! Decrease reference count
		/*! Destroys the object when reference count reaches zero.
		*/
		void downcount(void) const  
		{ 
			if (decrementRefCount(&crefs)== 0) 
			{
				delete this;	
			}
		}
		//! Dump member function
		void dump(Int32 level, const char *fmt, ...);

		//! Subsystem
		DPS_SUBSYSTEM(baseSubsystem);
	};

	//! Registrar for DPS class factory
	/*! \ingroup internal
	*/
	class Registrar
	{
	protected:
		//! First registered element
		static RegistrarItem *head;
		//! Last registered element
		static RegistrarItem *tail;
		//! Number of registered elements
		static Size count;
		//! Flattened view of registrar
		static RegistrarItem **items;
	public:
		//! Add item to registrar
		static void addItem(RegistrarItem *it)
		{
			for(RegistrarItem *c=head;c;c=c->next)
			{
				if(!strcmp(c->name,it->name))
					return;
			}
			if(tail)
				tail->next=it;
			else
				head=it;
			it->next=NULL;
			it->index=count;
			tail=it;
			count++;
		}

		//! Generate flat view
		static void finalize()
		{
			items=(RegistrarItem**)malloc(count*sizeof(RegistrarItem*));
			Size cnt=0;
			for(RegistrarItem *c=head;c;c=c->next)
				items[cnt++]=c;
		}

		//! Delete flat view
		static void destroy()
		{
			free(items);
		}

		//! Create an object of a named type
		static Ptr<Object> create(const AnsiChar *cname)
		{
			// Find the correct type in the chain of registered types
			for(RegistrarItem *c=head;c;c=c->next)
			{
				if(!strcmp(c->name,cname))
					return c->create();
			}
			return NULL;
		}

		//! Create an object of an indexed type
		static Ptr<Object> create(Size index)
		{
			return items[index]->create();
		}

		//! Get number of registered types
		static Size size()
		{
			return count;
		}

		//! Get index of a named type
		static Size indexOf(const AnsiChar *cname)
		{
			// Find the correct type in the chain of registered types
			for(Size i=0;i<count;i++)
			{
				if(!strcmp(items[i]->name,cname))
					return i;
			}
			return (Size)-1;
		}

		//! Get the name of type at index i
		static const AnsiChar *getName(Size i)
		{
			return items[i]->name;
		}
	};

//! Register a type under its type name
/*! \ingroup user
	If for some reason the internal registration mechanism should fail (i.e. the
	class factory cannot create an object of a given type, this macro can be used 
	in the global scope in order to force registration of a class.
*/
#define USETYPE(cname)			static base::RegistrarItemSpec<cname > registeritem##__LINE__

	//! Threaded object base class
	/*! \ingroup internal
		Objects that require one thread for their execution can derive
		from this class.
	*/
	class ThreadClass : public Object
	{
		//! Thread for this object
		THREAD_DECL(innerThread,run);
		
	public:
		//! Constructor
		ThreadClass() : THREAD_CONS(innerThread) { }
		
		//! Virtual trhead function
		virtual Int32 run() = 0;
		//! Start thread
		Result start() { return innerThread.start(); }
		//! Stop thread
		Result stop(Bool fast=false) { return innerThread.stop(fast); }
		//! Wait for end of thread
		Result wait() { return innerThread.wait(); }
		//! Should the thread stop
		bool shouldStop() { return innerThread.shouldStop(); }

#ifdef DPS_PLATFORM_WIN32
		//! Get root fiber
		void *getRootFiber() { return innerThread.getRootFiber(); }
#endif
		IDENTIFYNOREG(ThreadClass);
	};

}


// Include container classes
#include "basecont.h"

// Include socket classes
#include "basesock.h"

// Include serialization classes
#include "baseserial.h"

#endif

