/*
    DPS - Dynamic Parallel Schedules
    Copyright (C) 2000-2003 Sebastian Gerlach

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
	DPS Layer 1 - Base library

	This file is part of the base namespace, containing all the basic 
	functionality used by DPS. It provides support for reflection, smart
	pointers, serialization, and containers. It also contains cross-platform
	abstractions for several useful system services, such as threads, sockets,
	and synchronization primitives.
	The classes in this namespace have no dependencies on other DPS classes.
*/

/*! \file basesock.h
	\brief Socket classes

	Contains classes used for abstraction of sockets
*/

#ifndef INCLUDED_DPS_SOCKHELP_H
#define INCLUDED_DPS_SOCKHELP_H

#ifdef DPS_PLATFORM_WIN32
typedef int socklen_t;

#else

#ifdef DPS_NO_SOCKLEN_T
typedef int socklen_t;
#endif

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
//#include <dlfcn.h>
#include <netdb.h>
#include <unistd.h>

typedef int SOCKET;
#define SOCKET_ERROR -1
#define INVALID_SOCKET -1
#define closesocket(x) close(x)
typedef timeval TIMEVAL;

#endif

#define MKADDR(a,b,c,d) (((a))+((b)<<8)+((c)<<16)+((d)<<24))


namespace base
{
	//! TCP/IP address
	/*! \ingroup internal
	*/
	class TCPIPAddress
	{
		SIMPLECLASSDEF(TCPIPAddress)
			//! IP Address
			/*! Stored in network byte order.
			*/
			ITEM(UInt32,address)
			//! TCP Port
			/*! Stored in network byte order.
			*/
			ITEM(UInt16,port)
		CLASSEND;

	public:
		//! Constructor
		TCPIPAddress(UInt32 address = INADDR_ANY, UInt16 port = 0)
		{
			this->address=address;
			this->port=port;
		}

		//! Equality operator
		bool operator==(const TCPIPAddress& o) const
		{
			return address==o.address && port==o.port;
		}
		//! Less than operator
		bool operator<(const TCPIPAddress& o) const
		{
			return address<o.address || (address==o.address && port<o.port);
		}
		//! Return string form
		std::string format() const;
		//! Is the address valid?
		Bool valid() const { return address!=INADDR_ANY && port!=0; }
	};

	//! Useful functions for sockets
	/*! \ingroup internal
	*/
	class SocketHelper
	{
		UInt32 localAddress;
		std::string localName;

	public:
		//! Constructor, initializes networking subsystem
		SocketHelper();
		//! Destructor, deinitializes networking subsystem
		~SocketHelper();
		
		//! Get local host address
		UInt32 getLocalAddress() const { return localAddress; }
		//! Get local host name
		const std::string& getLocalName() const { return localName; }
		//! Format IP address
		static std::string formatIPAddress(UInt32 addr);
		//! Resolve a name
		static TCPIPAddress resolve(const char *name, UInt16 defPort = 5005);
	};

	//! Instance of socket helper
	extern SocketHelper socketHelper;


	//! Wrapper for TCP client sockets
	/*! \ingroup internal
		This class does not manage the lifecycle of the underlying socket.
	*/
	class TCPSocket : public Object
	{
	public:
		//! Underlying socket
		SOCKET sock;
		//! Use TCP_NODELAY
		static bool noDelay;

	private:
		//! Remote host address
		TCPIPAddress address;
		//! Remote host name
		std::string name;
		
	public:
		//! Constructor
		TCPSocket();
		//! Destructor
		~TCPSocket();

		//! Is socket connected?
		Bool connected() { return sock!=SOCKET_ERROR; }

		//! Attach class to socket
		/*!	\param sock Socket to attach to
			\param addr Source address of this connection
		*/
		Result connect(SOCKET sock, sockaddr_in *addr = NULL);

		//! Connect to indicated addres
		/*!	\param ipa IP Address
			\param prt TCP port
			\param retries Number of retries
		*/
		Result connect(UInt32 ipa, UInt16 prt, UInt32 retries = 2);

		//! Connect to indicated address
		/*!	\param ad Hostname of target machine (or IP address in string form)
			\param prt TCP port
		*/
		Result connect(const AnsiChar *ad, UInt16 prt = 0);

		//! Connect to indicated address
		/*! \param addr Address to connect to
		*/
		Result connect(const TCPIPAddress& addr);

		//! Disconnect this socket
		Result disconnect();

		//! Set options for this socket
		void setOptions();
		
		//! Send data on socket (might be a partial send)
		/*! \return Number of bytes sent
		*/
		Int32 psend(const void *data, Int32 len);

		//! Send data on socket (complete send)
		/*! \return Number of bytes sent
		*/
		Int32 send(const void *data, Int32 len);

		//! Send data on socket (complete send)
		/*! \param b Buffer to send
			\return Number of bytes sent
		*/
		template<typename T> Int32 send(Buffer<T>& b) { return send(b,b.size()); }

		//! Receive data on socket (might be a partial receive)
		/*! \param data Data to send
			\param len Length of data in bytes
			\return Number of bytes received
		*/
		Int32 precv(void *data, Int32 len);

		//! Receive data on socket (complete receive)
		/*! \param data Data to send
			\param len Length of data in bytes
			\return Number of bytes received
		*/
		Int32 recv(void *data, Int32 len);

		//! Skip data on socket
		/*! \param len Number of bytes to skip
			\return Number of bytes skipped
		*/
		Int32 skip(Int32 len);
		
		//! Check for availability of data to receive
		Int32 available(int timeout);

		//! Get address of this socket
		const TCPIPAddress& getAddress() const { return address; }

		//! Get name of this socket (string representation of address)
		const char *getName() { return name.c_str(); }
		
		IDENTIFYNOREG(TCPSocket);
	};

	//! Interface for incoming connection notification
	/*! \ingroup internal
	*/
	class IIncomingConnection 
	{
	public:
		//! Notification for incoming connections
		/*! \param sock Socket for connection
		*/
		virtual void incomingConnection(TCPSocket& sock) = 0;
	};
	
	//! Wrapper for TCP listen sockets
	/*! \ingroup internal
		This class wraps a TCP listen socket, using a listen thread to process
		incoming connections.
	*/
	class TCPListenSocket : public Object
	{
		IDENTIFYNOREG(TCPListenSocket);

		//! Underlying socket
		SOCKET sock;

		//! Address
		TCPIPAddress address;
		//! Name
		std::string name;
		//! Notification target
		IIncomingConnection *trg;
		//! Listen thread
		THREAD_DECL(listenThread,run);
		
	public:
		//! Constructor
		/*! \param target Target for notifications
			\param port Port to listen on
			\param address Address to listen on
		*/
		TCPListenSocket(IIncomingConnection *target, UInt16 port = 0, UInt32 address = INADDR_ANY);

		//! Destructor
		~TCPListenSocket();

		//! Thread function for listen socket
		Int32 run();

		//! Start listening
		void start() { listenThread.start(); }
		//! Stop listening
		void stop() { listenThread.stop(); }
		
		//! Get address of listen socket
		const TCPIPAddress& getAddress() const { return address; }

		//! Get name of listen socket
		const char *getName() { return name.c_str(); }
	};
}

#endif

