/*
    DPS - Dynamic Parallel Schedules
    Copyright (C) 2000-2003 Sebastian Gerlach

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
	DPS Layer 1 - Base library

	This file is part of the base namespace, containing all the basic 
	functionality used by DPS. It provides support for reflection, smart
	pointers, serialization, and containers. It also contains cross-platform
	abstractions for several useful system services, such as threads, sockets,
	and synchronization primitives.
	The classes in this namespace have no dependencies on other DPS classes.
*/

/*!	\file basetypes.h
	\brief Type handler classes

	These classes provide type handlers for all basic types, and for STL containers.
*/

#ifndef INCLUDED_BASE_TYPES_H
#define INCLUDED_BASE_TYPES_H

#include <vector>
#include <deque>
#include <list>
#include <set>

namespace base
{
#ifdef DPS_NAMED_SERIALIZATION
#define ENTERITEM(x) s->enter(x)
#define LEAVEITEM(x) s->leave(x)
#else
#define ENTERITEM(x)
#define LEAVEITEM(x)
#endif


	class Object;

	//! Vector of references to objects
	/*! \ingroup internal
		This class is used to store all referenced objects when deserializing.
	*/
	class RefVector
	{
	private:
		//! Referenced objects
		Object **contents;
		//! Number of elements in vector
		Size curSize;
		//! Maximum size of vector
		Size maxSize;
		//! Size increment when maximum reached
		Size increment;
		
	public:
		//! Constructor
		RefVector()
		{
			curSize=1;
			maxSize=10;
			increment=10;
			contents=(Object**)malloc(maxSize*sizeof(Object*));
			// The first element is always the NULL pointer
			contents[0]=NULL;
		}
		
		//! Destructor
		~RefVector();

		//! Add an element to the vector
		/*! If the object is already in the vector, it is not added.
			/param el Object to add
			/retval Index of the object in the vector
		*/
		Size addElement(Object* el);

		//! Get current size of vector
		inline Size size() const { return curSize; }
		
		//! Get an element from vector
		Object* elementAt(Size i) const { return contents[i]; }
	};

	//! Vector of constant references to objects
	/*! \ingroup internal
		This class is used to store all referenced objects when serializing.
	*/
	class ConstRefVector
	{
	private:
		//! Referenced objects
		const Object * *contents;
		//! Number of elements in vector
		Size curSize;
		//! Maximum size of vector
		Size maxSize;
		//! Size increment when maximum reached
		Size increment;
		
	public:
		//! Constructor
		ConstRefVector()
		{
			curSize=1;
			maxSize=10;
			increment=10;
			contents=(const Object**)malloc(maxSize*sizeof(Object*));
			// The first element is always the NULL pointer
			contents[0]=NULL;
		}
		
		//! Destructor
		~ConstRefVector();

		//! Add an element to the vector
		/*! If the object is already in the vector, it is not added.
			/param el Object to add
			/retval Index of the object in the vector
		*/
		Size addElement(const Object* el);

		//! Get current size of vector
		inline Size size() const { return curSize; }
		
		//! Get an element from vector
		const Object* elementAt(Size i) const { return contents[i]; }
	};

	//! Base class for output streams
	/*! \ingroup internal
	*/
	class OutputStream
	{
	protected:
		//! Referenced objects
		ConstRefVector refs;

	public:
		//! Destructor
		/*! To remove gcc warning.
		*/
		virtual ~OutputStream() { }

		virtual void writeInt8(const Int8& item) = 0;
		virtual void writeUInt8(const UInt8& item) = 0;
		virtual void writeInt16(const Int16& item) = 0;
		virtual void writeUInt16(const UInt16& item) = 0;
		virtual void writeInt32(const Int32& item) = 0;
		virtual void writeUInt32(const UInt32& item) = 0;
		virtual void writeInt64(const Int64& item) = 0;
		virtual void writeUInt64(const UInt64& item) = 0;
		virtual void writeFloat(const Float& item) = 0;
		virtual void writeDouble(const Double& item) = 0;
		virtual void writeBool(const Bool& item) = 0;
		virtual void writeString(const std::string& item) = 0;

		// For high performance blocks of simple types
		virtual void writeArray(const void *data, const Size count, const Size size) = 0;

		// For named serialization
		virtual void enter(const char *name) { }
		virtual void leave(const char *name) { }

		void writeRef(const Object* item);

		virtual Bool write(const Object *item) = 0;
	};

	//! Base class for input streams
	/*! \ingroup internal
	*/
	class InputStream
	{
	protected:
		//! Referenced objects
		RefVector refs;

	public:
		//! Destructor
		/*! To remove gcc warning.
		*/
		virtual ~InputStream() { }

		virtual void readInt8(char& item) = 0;
		virtual void readUInt8(unsigned char& item) = 0;
		virtual void readInt16(Int16& item) = 0;
		virtual void readUInt16(UInt16& item) = 0;
		virtual void readInt32(Int32& item) = 0;
		virtual void readUInt32(UInt32& item) = 0;
		virtual void readInt64(Int64& item) = 0;
		virtual void readUInt64(UInt64& item) = 0;
		virtual void readFloat(Float& item) = 0;
		virtual void readDouble(Double& item) = 0;
		virtual void readBool(Bool& item) = 0;
		virtual void readString(std::string& item) = 0;

		// For high performance blocks of simple types
		virtual void readArray(void *data, const Size count, const Size size) = 0;

		// For named serialization
		virtual void enter(const char *name) { }
		virtual void leave(const char *name) { }

		Object *readRef();

		virtual Object *read() = 0;
	};

	//! Contains all type handler classes
	/*! This namespace contains all classes used to handle types that do not have
		built-in reflection capabilities. This includes all the basic types (int, float,
		etc.) and also the STL containers and utility classes.
	*/
	namespace typehandler
	{
//! Static concatenation of 2 strings
/*! \ingroup internal
*/
#define RETSTRCAT2(s1,s2) static char *tn = NULL; if(tn==NULL) { tn=(char*)malloc(strlen(s1)+strlen(s2)+1); \
	strcpy(tn,s1); strcat(tn,s2); \
	} return tn
//! Static concatenation of 3 strings
/*! \ingroup internal
*/
#define RETSTRCAT3(s1,s2,s3) static char *tn = NULL; if(tn==NULL) { tn=(char*)malloc(strlen(s1)+strlen(s2)+strlen(s3)+1); \
	strcpy(tn,s1); strcat(tn,s2); strcat(tn,s3); \
	} return tn
//! Static concatenation of 4 strings
/*! \ingroup internal
*/
#define RETSTRCAT4(s1,s2,s3,s4) static char *tn = NULL; if(tn==NULL) { tn=(char*)malloc(strlen(s1)+strlen(s2)+strlen(s3)+strlen(s4)+1); \
	strcpy(tn,s1); strcat(tn,s2); strcat(tn,s3); strcat(tn,s4); \
	} return tn
//! Static concatenation of 5 strings
/*! \ingroup internal
*/
#define RETSTRCAT5(s1,s2,s3,s4,s5) static char *tn = NULL; if(tn==NULL) { tn=(char*)malloc(strlen(s1)+strlen(s2)+strlen(s3)+strlen(s4)+strlen(s5)+1); \
	strcpy(tn,s1); strcat(tn,s2); strcat(tn,s3); strcat(tn,s4); strcat(tn,s5); \
	} return tn
//! Static concatenation of 7 strings
/*! \ingroup internal
*/
#define RETSTRCAT7(s1,s2,s3,s4,s5,s6,s7) static char *tn = NULL; if(tn==NULL) { tn=(char*)malloc(strlen(s1)+strlen(s2)+strlen(s3)+strlen(s4)+strlen(s5)+strlen(s6)+strlen(s7)+1); \
	strcpy(tn,s1); strcat(tn,s2); strcat(tn,s3); strcat(tn,s4); strcat(tn,s5); strcat(tn,s6); strcat(tn,s7); \
	} return tn
//! Static concatenation of 9 strings
/*! \ingroup internal
*/
#define RETSTRCAT9(s1,s2,s3,s4,s5,s6,s7,s8,s9) static char *tn = NULL; if(tn==NULL) { tn=(char*)malloc(strlen(s1)+strlen(s2)+strlen(s3)+strlen(s4)+strlen(s5)+strlen(s6)+strlen(s7)+strlen(s8)+strlen(s9)+1); \
	strcpy(tn,s1); strcat(tn,s2); strcat(tn,s3); strcat(tn,s4); strcat(tn,s5); strcat(tn,s6); strcat(tn,s7); strcat(tn,s8); strcat(tn,s9); \
	} return tn

		//! Temporary conversion of integer to string
		/*! \ingroup internal
		*/
		struct IntToStringConverter
		{
			char value[16];
			IntToStringConverter(int i) { sprintf(value,"%d",i); }
		};

		//! Generic type handler
		/*! \ingroup internal
		*/
		template<typename x> struct TypeHandler 
		{ 
			static inline const char *name() { return x::getTypeName(); } 
			static inline void write(OutputStream *s, const x& item) { item.write(s); }
			static inline void read(InputStream *s, x& item) { item.read(s); }
		};

// There is a type handler for const types, but they cannot be serialized
// or deserialized (obvious, since they are const, they cannot be read from
// a file...)
#ifdef DPS_COMPILER_GCC
#define ARRAYOFSIMPLETYPE(x) template<int size> struct TypeHandler<x [size]> {\
	static inline const char *name() { RETSTRCAT4(TypeHandler<x>::name(),"[",IntToStringConverter(size).value,"]"); } \
	static inline void write(OutputStream *s, const x (&items) [size]) { ENTERITEM("array"); s->writeArray(&items,size,sizeof(x)); LEAVEITEM("array"); } \
	static inline void read(InputStream *s, x (&items) [size]) { ENTERITEM("array"); s->readArray(&items,size,sizeof(x)); LEAVEITEM("array"); } \
}; \
template<int size> struct TypeHandler<x const [size]> {\
	static inline const char *name() { RETSTRCAT4(TypeHandler<x>::name(),"[",IntToStringConverter(size).value,"]"); } \
	static inline void write(OutputStream *s, const x (&items) [size]) { ENTERITEM("array"); s->writeArray(&items,size,sizeof(x)); LEAVEITEM("array"); } \
	static inline void read(InputStream *s, x (&items) [size]) { ENTERITEM("array"); s->readArray(&items,size,sizeof(x)); LEAVEITEM("array"); } \
}
#else
#define ARRAYOFSIMPLETYPE(x)
#endif
#define SIMPLETYPE(x) template<> struct TypeHandler<x> { \
	static inline const char *name() { return #x; } \
	static inline void write(OutputStream *s, const x& item) { s->write##x(item); } \
	static inline void read(InputStream *s, x& item) { s->read##x(item); } \
}; \
template<> struct TypeHandler<const x> { \
	static inline const char *name() { return "const "#x; } \
}; \
ARRAYOFSIMPLETYPE(x)

		//! Type handler for signed 8-bit integers
		SIMPLETYPE(Int8);
		//! Type handler for unsigned 8-bit integers
		SIMPLETYPE(UInt8);
		//! Type handler for signed 16-bit integers
		SIMPLETYPE(Int16);
		//! Type handler for unsigned 16-bit integers
		SIMPLETYPE(UInt16);
		//! Type handler for signed 32-bit integers
		SIMPLETYPE(Int32);
		//! Type handler for unsigned 32-bit integers
		SIMPLETYPE(UInt32);
		//! Type handler for signed 64-bit integers
		SIMPLETYPE(Int64);
		//! Type handler for unsigned 64-bit integers
		SIMPLETYPE(UInt64);
		//! Type handler for 32-bit floating point numbers
		SIMPLETYPE(Float);
		//! Type handler for 64-bit floating point numbers
		SIMPLETYPE(Double);
		//! Type handler for booleans
		SIMPLETYPE(Bool);

		//! Type handler for STL strings
		template<> struct TypeHandler<std::string> 
		{ 
			static inline const char *name() { return "std::string"; } 
			static inline void write(OutputStream *s, const std::string& item) { s->writeString(item); } 
			static inline void read(InputStream *s, std::string& item) { s->readString(item); } 
		};
#ifdef DPS_COMPILER_GCC

		//! Type handler for STL allocator
		template<typename x> struct TypeHandler<std::allocator<x> > 
		{
			static inline const char *name() { RETSTRCAT3("std::allocator<",TypeHandler<x>::name(),">"); } 
		};

		//! Type handler for STL less
		template<typename x> struct TypeHandler<std::less<x> > 
		{
			static inline const char *name() { RETSTRCAT3("std::less<",TypeHandler<x>::name(),">"); } 
		};

		//! Type handler for STL pair
		template<typename a, typename b> struct TypeHandler<std::pair<a,b> > 
		{ 
			static inline const char *name() { RETSTRCAT5("std::pair<",TypeHandler<a>::name(),",",TypeHandler<b>::name(),">"); } 
			static inline void write(OutputStream *s, const std::pair<a,b>& item) 
			{ 
				ENTERITEM("first");
				typehandler::write(s,item.first);
				LEAVEITEM("first");
				ENTERITEM("second");
				typehandler::write(s,item.second);
				LEAVEITEM("second");
			}
			static inline void read(InputStream *s, std::pair<a,b>& item) 
			{ 
				ENTERITEM("first");
				typehandler::read(s,item.first);
				LEAVEITEM("first");
				ENTERITEM("second");
				typehandler::read(s,item.second);
				LEAVEITEM("second");
			} 
		};

		//! Type handler for STL vector
		template<typename x, typename a> struct TypeHandler<std::vector<x,a> > 
		{ 
			static inline const char *name() { RETSTRCAT5("std::vector<",TypeHandler<x>::name(),",",TypeHandler<a>::name(),">"); } 
			static inline void write(OutputStream *s, const std::vector<x,a>& item) 
			{ 
				ENTERITEM("size");
				XPSize size=(XPSize)item.size();
				typehandler::write(s,size);
				LEAVEITEM("size");
				for(typename std::vector<x,a>::const_iterator i=item.begin();i!=item.end();i++)
				{
					ENTERITEM("element");
					typehandler::write(s,*i);
					LEAVEITEM("element");
				}
			}
			static inline void read(InputStream *s, std::vector<x,a>& item) 
			{ 
				item.clear();
				ENTERITEM("size");
				XPSize size;
				typehandler::read(s,size);
				LEAVEITEM("size");
				for(Size i=0;i<size;i++)
				{
					ENTERITEM("element");
					x t;
					typehandler::read(s,t);
					item.push_back(t);
					LEAVEITEM("element");
				}
			} 
		};

		//! Type handler for STL deque
		template<typename x, typename a> struct TypeHandler<std::deque<x,a> > 
		{ 
			static inline const char *name() { RETSTRCAT5("std::deque<",TypeHandler<x>::name(),",",TypeHandler<a>::name(),">"); } 
			static inline void write(OutputStream *s, const std::deque<x,a>& item) 
			{ 
				ENTERITEM("size");
				XPSize size=(XPSize)item.size();
				typehandler::write(s,size);
				LEAVEITEM("size");
				for(typename std::deque<x,a>::const_iterator i=item.begin();i!=item.end();i++)
				{
					ENTERITEM("element");
					typehandler::write(s,*i);
					LEAVEITEM("element");
				}
			}
			static inline void read(InputStream *s, std::deque<x,a>& item) 
			{ 
				item.clear();
				ENTERITEM("size");
				XPSize size;
				typehandler::read(s,size);
				LEAVEITEM("size");
				for(Size i=0;i<size;i++)
				{
					ENTERITEM("element");
					x t;
					typehandler::read(s,t);
					item.push_back(t);
					LEAVEITEM("element");
				}
			} 
		};

		//! Type handler for STL list
		template<typename x, typename a> struct TypeHandler<std::list<x,a> > 
		{ 
			static inline const char *name() { RETSTRCAT5("std::list<",TypeHandler<x>::name(),",",TypeHandler<a>::name(),">"); } 
			static inline void write(OutputStream *s, const std::list<x,a>& item) 
			{ 
				ENTERITEM("size");
				XPSize size=(XPSize)item.size();
				typehandler::write(s,size);
				LEAVEITEM("size");
				for(typename std::list<x,a>::const_iterator i=item.begin();i!=item.end();i++)
				{
					ENTERITEM("element");
					typehandler::write(s,*i);
					LEAVEITEM("element");
				}
			}
			static inline void read(InputStream *s, std::list<x,a>& item) 
			{ 
				item.clear();
				ENTERITEM("size");
				XPSize size;
				typehandler::read(s,size);
				LEAVEITEM("size");
				for(Size i=0;i<size;i++)
				{
					ENTERITEM("element");
					x t;
					typehandler::read(s,t);
					item.push_back(t);
					LEAVEITEM("element");
				}
			} 
		};

		//! Type handler for STL set
		template<typename x, typename t, typename a> struct TypeHandler<std::set<x,t,a> > 
		{ 
			static inline const char *name() { RETSTRCAT7("std::set<",TypeHandler<x>::name(),",",TypeHandler<t>::name(),",",TypeHandler<a>::name(),">"); } 
			static inline void write(OutputStream *s, const std::set<x,t,a>& item) 
			{ 
				ENTERITEM("size");
				XPSize size=(XPSize)item.size();
				typehandler::write(s,size);
				LEAVEITEM("size");
				for(typename std::set<x,t,a>::const_iterator i=item.begin();i!=item.end();i++)
				{
					ENTERITEM("element");
					typehandler::write(s,*i);
					LEAVEITEM("element");
				}
			}
			static inline void read(InputStream *s, std::set<x,t,a>& item) 
			{ 
				item.clear();
				ENTERITEM("size");
				XPSize size;
				typehandler::read(s,size);
				LEAVEITEM("size");
				for(Size i=0;i<size;i++)
				{
					ENTERITEM("element");
					x tm;
					typehandler::read(s,tm);
					item.insert(tm);
					LEAVEITEM("element");
				}
			} 
		};

		//! Type handler for STL multiset
		template<typename x, typename t, typename a> struct TypeHandler<std::multiset<x,t,a> > 
		{ 
			static inline const char *name() { RETSTRCAT7("std::multiset<",TypeHandler<x>::name(),",",TypeHandler<t>::name(),",",TypeHandler<a>::name(),">"); } 
			static inline void write(OutputStream *s, const std::multiset<x,t,a>& item) 
			{ 
				ENTERITEM("size");
				XPSize size=(XPSize)item.size();
				typehandler::write(s,size);
				LEAVEITEM("size");
				for(typename std::multiset<x,t,a>::const_iterator i=item.begin();i!=item.end();i++)
				{
					ENTERITEM("element");
					typehandler::write(s,*i);
					LEAVEITEM("element");
				}
			}
			static inline void read(InputStream *s, std::multiset<x,t,a>& item) 
			{ 
				item.clear();
				ENTERITEM("size");
				XPSize size;
				typehandler::read(s,size);
				LEAVEITEM("size");
				for(Size i=0;i<size;i++)
				{
					ENTERITEM("element");
					x tm;
					typehandler::read(s,tm);
					item.insert(tm);
					LEAVEITEM("element");
				}
			} 
		};

		//! Type handler for STL map
		template<typename x, typename y, typename t, typename a> struct TypeHandler<std::map<x,y,t,a> > 
		{ 
			static inline const char *name() { RETSTRCAT9("std::map<",TypeHandler<x>::name(),",",TypeHandler<y>::name(),",",TypeHandler<t>::name(),",",TypeHandler<a>::name(),">"); } 
			static inline void write(OutputStream *s, const std::map<x,y,t,a>& item) 
			{ 
				ENTERITEM("size");
				XPSize size=(XPSize)item.size();
				typehandler::write(s,size);
				LEAVEITEM("size");
				for(typename std::map<x,y,t,a>::const_iterator i=item.begin();i!=item.end();i++)
				{
					ENTERITEM("element");
					typehandler::write(s,*i);
					LEAVEITEM("element");
				}
			}
			static inline void read(InputStream *s, std::map<x,y,t,a>& item) 
			{ 
				item.clear();
				ENTERITEM("size");
				XPSize size;
				typehandler::read(s,size);
				LEAVEITEM("size");
				for(Size i=0;i<size;i++)
				{
					ENTERITEM("element");
					std::pair<x,y> tm;
					typehandler::read(s,tm);
					item.insert(tm);
					LEAVEITEM("element");
				}
			} 
		};

		//! Type handler for STL multimap
		template<typename x, typename y, typename t, typename a> struct TypeHandler<std::multimap<x,y,t,a> > 
		{ 
			static inline const char *name() { RETSTRCAT9("std::multimap<",TypeHandler<x>::name(),",",TypeHandler<y>::name(),",",TypeHandler<t>::name(),",",TypeHandler<a>::name(),">"); } 
			static inline void write(OutputStream *s, const std::multimap<x,y,t,a>& item) 
			{ 
				ENTERITEM("size");
				XPSize size=(XPSize)item.size();
				typehandler::write(s,size);
				LEAVEITEM("size");
				for(typename std::multimap<x,y,t,a>::const_iterator i=item.begin();i!=item.end();i++)
				{
					ENTERITEM("element");
					typehandler::write(s,*i);
					LEAVEITEM("element");
				}
			}
			static inline void read(InputStream *s, std::multimap<x,y,t,a>& item) 
			{ 
				item.clear();
				ENTERITEM("size");
				XPSize size;
				typehandler::read(s,size);
				LEAVEITEM("size");
				for(Size i=0;i<size;i++)
				{
					ENTERITEM("element");
					std::pair<x,y> tm;
					typehandler::read(s,tm);
					item.insert(tm);
					LEAVEITEM("element");
				}
			} 
		};


		//! Type handler for fixed-size arrays
		template<typename x, int size> struct TypeHandler<x [size]>
		{
			static inline const char *name() { RETSTRCAT4(TypeHandler<x>::name(),"[",IntToStringConverter(size).value,"]"); } 
			static inline void write(OutputStream *s, const x (&items) [size]) 
			{
				for(Size i=0;i<size;i++)
				{
					ENTERITEM("element");
					typehandler::write(s,items[i]);
					LEAVEITEM("element");
				}
			}
			static inline void read(InputStream *s, x (&items) [size]) 
			{ 
				for(Size i=0;i<size;i++)
				{
					ENTERITEM("element");
					typehandler::read(s,items[i]);
					LEAVEITEM("element");
				}
			} 
		};

#endif

		//! Write an element to the output stream
		/*! \param s Target output stream
			\param item Item to write
			This function passes the item to the appropriate type handler. It is 
			used to take advantage of the automatic template parameter selection
			feature of templated functions.
		*/
		template<typename x> void write(OutputStream *s, const x& item)
		{
			TypeHandler<x>::write(s,item);
		}

		//! Read an element from the input stream
		/*! \param s Source input stream
			\param item Item to read
			This function passes the item to the appropriate type handler. It is 
			used to take advantage of the automatic template parameter selection
			feature of templated functions.
		*/
		template<typename x> void read(InputStream *s, x& item)
		{
			TypeHandler<x>::read(s,item);
		}
	}

	//! Write a reference to the stream
	inline void OutputStream::writeRef(const Object* item)
	{
		XPSize ndx=(XPSize)refs.addElement(item); 
		typehandler::write(this,ndx);
	}

	//! Read a reference from the stream
	inline Object * InputStream::readRef()
	{
		XPSize ndx;
		typehandler::read(this,ndx);
		return refs.elementAt(ndx);
	}
}

#endif

